/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "DataSource.hh"
#include "rndm.hh"
#include <sstream>
#include <iostream>
#include <iomanip>
#include <math.h>

using namespace std;

namespace generator {

  //====================================  Source constructor
  DataSource::DataSource(RepMode mode) 
    : mRptMode(mode), mRate(0.0),mSample(1./16384.), mFlags(0),
      mNextTrig(0), mCurrent(0), mLatest(0), mEventNo(0), mSumTime(0)
  {
  }
  
  //====================================  Source destructor
  DataSource::~DataSource(void) {
  }

  //====================================  Dump the source status
  ostream&
  DataSource::dump(ostream& out) const {
      out << "Dump of data source: " << getName() 
	  << " type: " << getDataType() << endl;
      out << "Timing: ";
      if (mRptMode == kContinuous) {
	  out << "Continuous";
      } else if (mRptMode == kSingle) {
	  out << "Single trigger at " << mNextTrig;
      } else if (mRptMode == kRate) {
	  out << "Average rate = " << mRate;
      } else if (mRptMode == kFixed) {
	  out << "Fixed step of " << mRate;
      }
      out << endl;
      out << "Parameters: " << endl;
      for (const_param_iter i=mParam.begin() ; i != mParam.end(); ++i) {
	  out << "   " << i->first << ":  ";
	  i->second.dump(out);
      }
      return out;
  }

  //====================================  Dump the source defintion string
  ostream&
  DataSource::dumpDefinition(ostream& out) const {
      out << getDataType() << "(";
      for (const_param_iter i=mParam.begin() ; i != mParam.end(); ++i) {
	  if (i != mParam.begin()) out << ",";
	  out << i->first << "=";
	  i->second.dumpDefinition(out);
      }
      out << ")";
      return out;
  }

  //====================================  Set poisson repeat mode
  void 
  DataSource::setTriggerRate(double rate) {
      mRate    = rate;
      mRptMode = kRate;
  }

  //====================================  Set single event mode
  void 
  DataSource::setTriggerTime(const Time& t0) {
      mCurrent = t0;
      mRptMode = kSingle;
  }

  //====================================  Get the data series
  TSeries 
  DataSource::getTimeSeries(const Time& t0, Interval dT) const {
      return mSeries.extract(t0, dT);
  }

  //====================================  Get a string parameter value
  const char* 
  DataSource::getParameter(const string& name) const {
      const_param_iter p = findParam(name);
      if (p == mParam.end()) return reinterpret_cast<const char*>(0);
      return p->second.getString();
  }

  //====================================  Get a numeric parameter value
  double      
  DataSource::getNumeric(const string& name) const {
      const_param_iter p = findParam(name);
      if (p == mParam.end()) return 0.0;
      return p->second.getNumeric();
  }

  //====================================  Set a parameter
  void 
  DataSource::setParameter(const string& name, const string& value) {
      if (value.find("(") == value.npos) { 
	  mParam[name].setParameter(value);
      } else { 
	  mParam[name].setDistribution(value);
      }
  }

  //====================================  Set a parameter
  void 
  DataSource::setParameter(const string& name, double value) {
      mParam[name].setParameter(value);
  }

  //====================================  Generate the appropriate TSeries
  void 
  DataSource::generate(const Time& t0, Interval dt) {
      if (mDebug) {
	  cout << "Starting generation of " << getSourceName() << " for GPS "
	       << t0.getS() << " (length " << dt << ")." << endl;
      }

      if (mSeries.isEmpty() || mSeries.getEndTime() <= t0 ) {
	  float x(0.0);
	  mSeries = TSeries(t0, getSample(), 1, &x);
      }
      Time tEnd = t0 + dt;
      mSeries.extend(tEnd);

      //--------------------------------  Get first event parameters.
      if ((mRptMode == kRate || mRptMode == kFixed) && !mCurrent) {
	  mCurrent = t0;
      }
      randomize();
      if (isContinuous()) {
	  mTrigStart = t0;
	  mTrigEnd   = tEnd;
      }

      //--------------------------------  Loop while there's
      while (mTrigStart < tEnd && (mNextTrig != Time(0) || isContinuous())) {
	  if (mDebug > 1) {
	      cout << "   Generating ";
	      if (!isContinuous()) cout << getSourceName() << " ";
	      cout << "for GPS " << mTrigStart << " - " << mTrigEnd << endl;

	  }
	  mSeries.extend(mTrigEnd);
	  float* pF = reinterpret_cast<float*>(mSeries.refData());
	  int inx0 = mSeries.getBin(t0);
	  int len  = mSeries.getBin(mTrigEnd) - inx0;
	  getSeries(t0, getSample(), len, pF);
	  if (getSaveFlag()) {
	      ostringstream name;
	      name << mName << setw(8) << setfill('0') << mEventNo;
	      mRecord.push_back(SrcEvent(name.str(), getSourceName(), 
					 mCurrent, mCurrent-mTrigStart, 
					 mTrigEnd-mCurrent, mParam)
				);
	  }
	  mNextTrig = Time(0);
	  if (mRptMode == kSingle) mCurrent = Time(0);
	  mSumTime += mTrigEnd - mTrigStart;
	  mTrigStart = mTrigEnd;
	  randomize();
      }
      if (mDebug) cout << getSourceName() << " generation complete for GPS "
		       << t0.getS() << " - " << tEnd.getS() << ", " 
		       << mSeries.getNSample() << " samples." << endl;
      mLatest = tEnd;
  }

  //==================================== Select event time & set parameters
  ostream& 
  DataSource::print_stats(ostream& out) const {
      out << "Data Source: " << getName() 
	  << " Type: " << getDataType() << endl;
      out << "Source definition:          "; dumpDefinition(out) << endl;
      out << "Number of events generated: " << mEventNo << endl;
      out << "Total time:                 " << mSumTime << endl;
      return out;
  }

  //==================================== Select event time & set parameters
  void 
  DataSource::randomize(void) {
      if (!mNextTrig) {

	  //----------------------------  Generate a time
	  switch (mRptMode) {
	  case kRate: {
	      mNextTrig = mCurrent;
	      double z  = Rndm();
	      if (z <= 0) mNextTrig += Interval(999./mRate);
	      else        mNextTrig += -log(z)/mRate;
	      mCurrent = mNextTrig;
	      break;
	  }

	  case kFixed:
	      mNextTrig = mCurrent + Interval(1.0/mRate);
	      mCurrent = mNextTrig;
	      break;

	  case kSingle:
	      mNextTrig = mCurrent;
	      break;

	  default:
	      mCurrent = mNextTrig;
	  }

	  //----------------------------  Generate parameters
	  if (mNextTrig != Time(0)) {
 	      bool step = true;
	      for (param_iter i=mParam.begin() ; i != mParam.end() ; ++i) {
		  if (i->second.isConst()) {
		      continue;
		  } else if (! i->second.isStep()) {
		      i->second.sampleParameter();
		  } else if (step) {
		      i->second.sampleParameter();
		      step = i->second.atStart();
		  }
	      }
	      mTrigStart = mNextTrig  + getStartOffset();
	      mTrigEnd   = mTrigStart + getDuration(); 
	      mEventNo++;
	  }
      }
  }

  //====================================  Set the sample rate
  void 
  DataSource::release(const Time& t0) {
      Time tBegin = mSeries.getStartTime();
      if (tBegin < t0) mSeries.eraseStart(t0 - tBegin);
  }

  //====================================  Set the sample rate
  void 
  DataSource::setSample(Interval dT) {
      mSample = dT;
  }

  //====================================  Set the debuggin printout level
  void 
  DataSource::setDebug(int lvl) {
      mDebug = lvl;
  }

  //====================================  Set the source name
  void 
  DataSource::setName(const string& name) {
      mName = name;
  }

  //====================================  Set the save flag
  void 
  DataSource::setSaveFlag(bool yorn) {
      if (yorn) mFlags |=  flSvEvents;
      else      mFlags &= ~flSvEvents;
  }

  //====================================  Set the write flag
  void 
  DataSource::setWriteFlag(bool yorn) {
      if (yorn) mFlags |=  flWrite;
      else      mFlags &= ~flWrite;
  }

  //====================================  Get the source decription
  string
  DataSource::getSourceName(void) const {
      ostringstream ss;
      ss << getDataType();
      if (!mName.empty()) ss << " " << mName;
      ss << "(";
      for (const_param_iter i=mParam.begin() ; i != mParam.end() ; ++i) {
	  if (i != mParam.begin()) ss << ", ";
	  ss << i->first << "=" << i->second.getString();
      }
      ss << ")";
      return ss.str();
  }

  DataSource::SrcEvent::SrcEvent(void) {
  }

  DataSource::SrcEvent::SrcEvent(const std::string& name, 
				 const std::string& comment, 
				 const Time& t0, Interval before, 
				 Interval after, const ParamMap& map) 
    : mName(name), mComment(comment), mGPSMax(t0), 
      mTimeBefore(before), mTimeAfter(after), mAmplitude(0.0)
  {
      for (DataSource::const_param_iter i=map.begin(); i != map.end() ; i++) {
	  if (i->first == "A") mAmplitude = i->second.getNumeric();
	  mParamList.push_back(ParamEntry(i->first, i->second.getNumeric()));
      }
  }

}  // namespace generator
