/* -*- mode: c++; c-basic-offset: 3; -*- */
// $Id$
#ifndef WPIPE_WEVENNTLIST_HH
#define WPIPE_WEVENTLIST_HH

#include "wtypes.hh"
#include "Time.hh"
#include <string>
#include <iosfwd>

namespace wpipe {

   class wtransform;
   class qTransform;
   class wtile;
   class wouttype;

   /**  The wevent structure contains all the event properties used in the 
    *  various Omega ppeline event manipulation functions, such as:
    *  wthreshold, wcluster, wselect, wveto, etc.
    */
   struct wevent {
      /**  Contruct an empty event.
       */
      wevent(void);
    
      /** Compare two events according to normalized energy.
       */
      bool operator<(const wevent& evt) const;

      /** Compare two events according to normalized energy.
       */
      bool operator>(const wevent& evt) const;

      /**  Display the contents of an event.
       */
      void dump(std::ostream& out) const;

      double t_offset;           ///< Center time offset of the event
      double frequency;          ///< Central frequency of the event.
      double q;                  ///< Central Q value of the event.
      double duration;           ///< Event duration.
      double bandwidth;          ///< Eventband-width
      double normalizedEnergy;   ///< Normalized event energy
      double amplitude;          ///< Event amplitude.
      double incoherentEnergy;   ///< Event incoherent energy
      size_t cluster_size;       ///< Number of tiles composing event
      size_t cluster_id;         ///< ID of cluster this event belongs to.
   };

   //====================================  wevent inline functions.
   inline bool 
   wevent::operator<(const wevent& evt) const {
      return normalizedEnergy < evt.normalizedEnergy;
   }

   inline bool 
   wevent::operator>(const wevent& evt) const {
      return normalizedEnergy > evt.normalizedEnergy;
   }

   /**  The weventlist class is an attempt to unify the Omega pipeline event
    *  manipulation functions in a single class. The wevent structure
    *  contains the basic 
    */
   class weventlist {
   public:
      /**  Construct an empty event list with a specified class name. The
       *  event list class is optional and is used only in pronting out 
       *  the lists.
       */
      weventlist(const std::string& list_class="");

      /**  Add an event to the list.
       */
      void addEvent(const wevent& evt);

      /**  Return the channel name for which this list was created.
       */
      const std::string& channelName(void) const;

      /**  wcluster identifies clusters of related significant Q transform 
        *  tiles to improve the detection of gravitational-wave bursts that 
	*  are extended in time and/or frequency.
	*
	*  wcluster permits two choices of clustering methods: density and 
	*  hierarchical. The input arguments depends upon which method is 
	*  requested. 
	*  The current event list instance is filled with the clusters
	*  found and the argument event list is updated to reflect the
	*  cluster number to which each original event was assigned.
	*
	*  <b> Density based clustering <\b>
	*
	*  Density based clustering first identifies the number of tiles within
	*  a specified clustering radius of a given tile. If the number of 
	*  tiles within this radius meets a threshold cluster density, all of  
	*  the tiles within this radius become part of the same cluster.  
	*  Clusters are constructed by recursively applying this same test to 
	*  all of the other tiles within the clustering radius, until a 
	*  complete set of clusters have been formed.  By default, both 
	*  clusterRadius and clusterDensity are set to unity.
	*
	*  Density based clustering does not in general assign all tiles to 
	*  clusters, but if (the cluster singles argument is non zero, single
	*  tiles are included in the output cluster list.  Otherwise, single
	*  tiles are not included, and zeros are reported for their cluster 
	*  properties in the output significants structure.  By default,
	*  clusterSingles is false.
	*
	*  The significant tile structure is updated to contain the cluster 
	*  identification number and number of tiles in cluster
	*  \brief Density -based cluster finding.
	*  \param significants   Event list of significant tile properties
	*  \param distanceMetric Choice of metric to compute distance (see 
	*                        distance())
	*  \param durationInflation Multiplicative scale factor for duration 
	*                           (see distance())
	*  \param bandwidthInflation Multiplicative scale factor for bandwidth
	*                           (see distance())
	*  \param debugLevel Verbosity of debug output
	*/
      void wcluster(weventlist& significants, double clusterRadius=1.0, 
		    double clusterDensity=1.0, bool clusterSingles=false, 
		    const std::string& distanceMetric="",
		    double durationInflation=1.0, 
		    double bandwidthInflation=1.0, 
		    int debugLevel=1);


      /**  wcluster identifies clusters of related significant Q transform 
        *  tiles to improve the detection of gravitational-wave bursts that are
	*  extended in time and/or frequency.
	*
	*  <b> Hierarchical clustering <\b>
	*
	*  In hierarchical clustering, the distances returned by WDISTANCE are
	*  used by the linkage class to construct a hierarchical cluster tree 
	*  by successively linking together the next closest pair of tiles or 
	*  existing clusters.  Finally, clusters are identified by cutting the
	*  hieararchical cluster tree based on the specified criterion and
	*  threshold.
	*
	*  The cluster linkage method defines how intercluster distance is 
	*  computed, and may be any of the following methods.  By default,
	*  'single' linkage is assumed.
	*
	*  <table>
	*  <tr><td>'single'</td>
	*      <td> nearest distance between cluster members</td></tr>
	*  <tr><td>'complete'</td>
	*      <td>furthest distance between cluster members</td></tr>
	*  <tr><td>'average'</td>
	*      <td>average distance between cluster members</td></tr>
	*  <tr><td>'centroid'</td>
	*      <td>center of mass distance between clusters</td></tr>
	*  <tr><td>'ward'</td>
	*      <td>inner squared distance between clusters</td></tr>
	*  </table>
	*
	*  The cluster criterion specifies how to cut the hierarchical cluster 
	*  tree to form clusters, and may be either of the two methods 
	*  understood by the CLUSTER function.  By deafult, 'inconsistent' is 
	*  assumed.
	*  <table>
	*  <tr><td>'inconsistent'</td>
	*      <td>threshold on inconsistency between of cluster distance and
	*          average subcluster difference</td></tr>
	*  <tr><td>'distance'</td>
	*      <td>threshold on absolute distance between clusters</td></tr>
	*  </table>
	*
	*  The clusterThreshold corresponds to the specified clusterCriterion, 
	*  and has a default threshold of unity for both criteria.
	*/
      void wcluster(weventlist& significants, 
		    const std::string& clusterLinkage, 
		    const std::string& clusterCriterion, 
		    double clusterThreshold, 
		    const std::string& distanceMetric,
		    double durationInflation, double bandwidthInflation, 
		    int debugLevel);

      void dump(std::ostream& out) const;
      const std::string& eventClass(void) const;
      void limit(size_t maxEvents);
      const Time& refTime(void) const;
      size_t size(void) const;
      const wevent& operator[](size_t inx) const;
      wevent& operator[](size_t inx);
      void sort(void);

      /** WSELECT selects statistically significant events from the set of 
        * statistically significant Q transform tiles. Events are defined by 
	* the properties of their most significant tile and are identified 
	* by exluding the less significant of any overlapping tiles.  The 
	* input significant tiles are first sorted by decreasing normalized 
	* energy.  Starting with the most significant tile, tiles are discarded
	* if (they overlap with a more significant tile.  The remaining set of 
	* tiles comprises a minimal set of tiles that describes an event.
	* the results are left in the cuurent eventlist instance.
	*
	* The optional durationInflation and bandwidthInflation arguments are
	* multiplicative scale factors that are applied to the duration and 
	* bandwidth of significant tiles prior to testing for overlap.  If not 
	* specified, these parameters both default to unity such that the 
	* resulting tiles have unity time-frequency area.  The normalized 
	* energy of the resulting tiles are scaled by the product of the 
	* duration and bandwidth inflation factors to avoid over counting the 
	* total energy of clusters of tiles.  Likewise, the amplitude of the 
	* resulting tiles is scaled by the square root of the product of the
	* duration and bandwidth inflation factors.
	*
	* The optional maximumEvents argument provides a safety mechanism to 
	* limit the total number of events returned by WSELECT.  If this 
	* maximum number of events is exceeded, an overflow flag is set, only 
	* the maximumEvents most significant events are returned, and a warning
	* is issued if (debugLevel is set to unity or higher.  By default, 
	* maximumEvents is set to infinity and debugLevel is set to unity.
	*
	* WSELECT both expects and returns a cell array of Q transform event 
	* structures with one cell per channel.  The event structures contain 
	* the following required fields, which describe the properties of 
	* statistically significant tiles.  Additional fields such as  
	* amlitude phase, or coherent transform properties are optional and 
	* are retained along with the required fields.
	*
	*   time                 center time of tile [gps seconds]
	*   frequency            center frequency of tile [Hz]
	*   duration             duration of tile [seconds]
	*   bandwidth            bandwidth of tile [Hz]
	*   normalizedEnergy     normalized energy of tile []
	*
	* The event list also contains the following flag, which indicates if 
	* the maximum number of significant tiles or significant events was 
	* exceeded.
	*
	* \param significants      event list tiles properties. This is assumed
	*                          to be ordered by decreasing normalizedEnergy.
	* \param durationInflation multiplicative scale factor for duration
	* \param bandwidthInflation multiplicative scale factor for bandwidth
	* \param maximumEvents     maximum allowable number of events
	* \param debugLevel        verbosity of debug output
	*/
      void wselect(const weventlist& significants, double durationInflation=1,
		   double bandwidthInflation=1, size_t maximumEvents=0, 
		   int debugLevel=1);

      /**  wthreshold identifies discrete Q transform coefficients whose 
       *  magnitudes exceed the threshold that approximately yields the 
       *  specified single channel false rate assuming ideal white noise.
       *  Each such tile is used to create an event which is added to the 
       *  list.
       *
       *  The user can focus on a subset of the times and frequencies available
       *  in the transform data by specifying a desired range of central times,
       *  central frequencies, and Qs to threshold on.  Ranges should be 
       *  specified as a two component vector, consisting of a minimum and 
       *  maximum value. Alternatively, if (only a single Q is specified, 
       *  WTHRESHOLD is only applied to the time-frequency plane which has the 
       *  nearest value of Q in a logarithmic sense to the requested value.
       *
       *  To determine the range of central times to threshold on, WTHRESHOLD
       *  requires the start time of the transformed data in addition to a
       *  reference time and a relative time range.  Both the start time and
       *  reference time should be specified as absolute quantities, while the
       *  range of times to analyze should be specified relative to the 
       *  requested reference time.
       *
       *  By default, WTHRESHOLD is applied to all available frequencies and Qs,
       *  and the reference time and relative time range arguments are set to
       *  exclude data potentially corrupted by filter transients as identified 
       *  by the transient duration field of the tiling structure.  The default 
       *  value can be obtained for any argument by passing an empty vector.
       *
       *  The threshold is set to yield the specified false event rate when 
       *  applied to all available frequencies and Qs, and is not modified 
       *  to account for restricted ranges.  It is also only a rough estimate,
       *  and the result false event rate may vary significantly depending on 
       *  the quality of the data.
       *
       *  If provided, the optional analysisMode string is used by WTHRESHOLD 
       *  to determine which channels are signal channels, which channels are 
       *  null channels, and which channels to report results for.  For 
       *  coherent analysis modes, a desired white noise veto rate, squared 
       *  calibration uncertainy factor, and required signal correlation factor
       *  must also be specified.
       *
       *  The optional maximumSignificants argument provides a safety mechanism 
       *  to limit the total number of events returned by WTHRESHOLD.  If this 
       *  maximum number of significants is exceeded, the overflow flag is set, 
       *  only the maximumSignificants most significant tiles are returned, and
       *  a warning is issued if (debugLevel is set to 1 or higher.  By default,
       *  maximumSignificants is set to 1e9-1 and debugLevel is set to unity.
       *
       *  \brief Generate list of significant tiles.
       *  \param signal         Input signal Q transform structure
       *  \param tiling         Discrete Q transform tiling structure from WTILE
       *  \param startTime      GPS start time of Q transformed data
       *  \param falseEventRate desired white noise false event rate [Hz]
       *  \param referenceTime  reference time for time range to threshold on
       *  \param timeRange      vector range of relative times to threshold on
       *  \param frequencyRange vector range of frequencies to threshold on
       *  \param qRange         scalar Q or vector range of Qs to threshold on
       *  \param maximumSignificants maximum allowable number of events
       *  \param falseVetoRate     desired white noise veto rate [Hz]
       *  \param uncertaintyFactor squared calibration uncertainty factor
       *  \param debugLevel        verbosity of debug output
       */
      void wthreshold(const qTransform& signal, const wtile& tiling, 
		      const Time& startTime, double falseEventRate, 
		      const Time& referenceTime, const dble_vect& timeRange, 
		      const dble_vect& frequencyRange, const dble_vect& qRange, 
		      size_t maximumSignificants, double uncertaintyFactor=0.0,
		      int debugLevel=1);

      /**  wthreshold identifies discrete Q transform coefficients whose 
        *  magnitudes exceed the threshold that approximately yields the 
	*  specified single channel false rate assuming ideal white noise.
	*  Each such tile is used to create an event which is added to the 
	*  list.
	*
	*  The user can focus on a subset of the times and frequencies 
	*  available in the transform data by specifying a desired range 
	*  of central times, central frequencies, and Qs to threshold on.
	*  Ranges should be specified as a two component vector, consisting
	*  of a minimum and maximum value. Alternatively, if (only a single
	*  Q is specified, WTHRESHOLD is only applied to the time-frequency
	*  plane which has the nearest value of Q in a logarithmic sense to
	*  the requested value.
	*
	*  To determine the range of central times to threshold on, WTHRESHOLD
	*  requires the start time of the transformed data in addition to a
	*  reference time and a relative time range.  Both the start time and
	*  reference time should be specified as absolute quantities, while the
	*  range of times to analyze should be specified relative to the 
	*  requested reference time.
	*
	*  By default, WTHRESHOLD is applied to all available frequencies and
	*  Qs, and the reference time and relative time range arguments are 
	*  set to exclude data potentially corrupted by filter transients as
	*  identified by the transient duration field of the tiling structure.
	*  The default value can be obtained for any argument by passing an 
	*  empty vector.
	*
	*  The threshold is set to yield the specified false event rate when 
	*  applied to all available frequencies and Qs, and is not modified 
	*  to account for restricted ranges.  It is also only a rough estimate,
	*  and the result false event rate may vary significantly depending on 
	*  the quality of the data.
	*
	*  If provided, the optional analysisMode string is used by WTHRESHOLD 
	*  to determine which channels are signal channels, which channels are 
	*  null channels, and which channels to report results for.  For 
	*  coherent analysis modes, a desired white noise veto rate, squared 
	*  calibration uncertainy factor, and required signal correlation 
	*  factor must also be specified.
	*
	*  The optional maximumSignificants argument provides a safety 
	*  mechanism to limit the total number of events returned by 
	*  WTHRESHOLD.  If this maximum number of significants is exceeded,
	*  the overflow flag is set, only the maximumSignificants most 
	*  significant tiles are returned, and a warning is issued if 
	*  debugLevel is set to 1 or higher. By default, maximumSignificants
	*  is set to 1e9-1 and debugLevel is set to unity.
	*
	*  \brief Generate list of significant tiles.
	*  \param signal         Input signal Q transform structure
	*  \param refer          Input null-stream Q transform structure
	*  \param tiling         Discrete Q transform tiling structure from 
	*                        WTILE
	*  \param startTime      GPS start time of Q transformed data
	*  \param falseEventRate desired white noise false event rate [Hz]
	*  \param referenceTime  reference time for time range to threshold on
	*  \param timeRange      vector range of relative times to threshold on
	*  \param frequencyRange vector range of frequencies to threshold on
	*  \param qRange         scalar Q or vector range of Qs to threshold on
	*  \param maximumSignificants maximum allowable number of events
	*  \param uncertaintyFactor squared calibration uncertainty factor
	*  \param correlationFactor fractional correlated energy threshold
	*  \param debugLevel        verbosity of debug output
       */
      void wthreshold(const qTransform& signal, const qTransform& refer, 
		      const wtile& tiling, const Time& startTime, 
		      double falseEventRate, const Time& referenceTime, 
		      const dble_vect& timeRange, 
		      const dble_vect& frequencyRange, const dble_vect& qRange,
		      size_t maximumSignificants, double uncertaintyFactor=0.0,
		      double correlationFactor=0.0, int debugLevel=1);

      /**  Write the specified fields of the event list to the output file.
       */
      void writeEvent(const std::string& path, const str_vect& fields, 
		      const std::string& format) const;

   private:
      /**  For both density and hierarchical clustering, distances between all 
        *  pairs of tiles are first computed using the WDISTANCE function and 
	*  the specified distance metric and inflation parameters.
	*
	*  The distance metric specified how distances are computed between 
	*  tiles, and may be any of the methods understood by the WDISTANCE 
	*  function.  By default, 'integratedMismatch' is assumed.
	*  <table>
	*  <tr><td>'pointMismatch'</td>
	*      <td>Second order approximation to mismatch function evaluated 
	*          at center point between tiles</td></tr>
	*  <tr><td>'integratedMismatch'</td>
	*      <td>Second order approximation to mismatch function integrated 
	*          between tiles</td></tr>
	*  <tr><td>'logMismatch'</td>
	*      <td>negative log of the overlap between two tiles</td></tr>
	*  <tr><td>'euclidean'</td>
	*      <td>normalized time frequency distance between tiles</td></tr>
	*  <tr><td>'modifiedEuclidean'</td>
	*      <td>normalized time frequency distance between tiles modified 
	*          to give greater weight to frequency</td></tr>
	* </table>
	*
	*  The specified duration and bandwidth inflation factors are used by 
	*  some distance metrics to compute the distance between tiles.  They 
	*  are also used when computing the total energies of clusters.  By 
	*  default, they are set to unity.
	*
	* \param distances Pointer to pre-allocated output array of dimension
	*                  N*(N-1)/2.
	* \param method Name of method to be used to calculate distances
	* \param durationInflation  Inflation factor for tile duration.
	* \param bandwidthInflation Inflation factor for tile bandwidth.
	*/
      void wdistance(double distances[], const std::string& method, 
		     double durationInflation=1.0, 
		     double bandwidthInflation=1.0);

      /**  Fill the event list with clustered events from the argument 
        *  list.
	*/
      void clusterFill(const weventlist& ev, double dtInflation,
		       double bwInflation);

   private:
      std::string _channelName;  ///< Channel name for this list
      std::string _list_class;   ///< Event class (e.g. raw, clistered, etc)
      int         _overflowFlag; ///< more than the maximum number of events.
      typedef std::vector<wevent> event_vect;
      event_vect  _events;       ///< list of events
      Time        _refTime;      ///< Reference time for all events in list.
      bool        _sorted;       ///< List is sorted by decreasing normE
   };

   //================================  weventlist inline accessors.
   inline const std::string& 
   weventlist::channelName(void) const {
      return _channelName;
   }

   inline const std::string& 
   weventlist::eventClass(void) const {
      return _list_class;
   }

   inline const Time&
   weventlist::refTime(void) const {
      return _refTime;
   }

   inline size_t 
   weventlist::size(void) const {
      return _events.size();
   }

   inline const wevent& 
   weventlist::operator[](size_t inx) const {
      return _events[inx];
   }

   inline wevent& 
   weventlist::operator[](size_t inx) {
      return _events[inx];
   }

   /**  The eventstack class contains a list of several event lists for multiple
    *  channels.
    */
   class weventstack {
   public:
      void dump(std::ostream& out) const;
      const weventlist& operator[](int i) const;
      int numberOfChannels(void) const;
      void status(std::ostream& out) const;


      /**  wcluster identifies clusters of related significant Q transform 
        *  tiles to improve the detection of gravitational-wave bursts that 
	*  are extended in time and/or frequency.
	*
	*  wcluster permits two choices of clustering methods: density and 
	*  hierarchical. The input arguments depends upon which method is 
	*  requested. 
	*  The current event list instance is filled with the clusters
	*  found and the argument event list is updated to reflect the
	*  cluster number to which each original event was assigned.
	*
	*  <b> Density based clustering <\b>
	*
	*  Density based clustering first identifies the number of tiles within
	*  a specified clustering radius of a given tile. If the number of 
	*  tiles within this radius meets a threshold cluster density, all of  
	*  the tiles within this radius become part of the same cluster.  
	*  Clusters are constructed by recursively applying this same test to 
	*  all of the other tiles within the clustering radius, until a 
	*  complete set of clusters have been formed.  By default, both 
	*  clusterRadius and clusterDensity are set to unity.
	*
	*  Density based clustering does not in general assign all tiles to 
	*  clusters, but if (the cluster singles argument is non zero, single
	*  tiles are included in the output cluster list.  Otherwise, single
	*  tiles are not included, and zeros are reported for their cluster 
	*  properties in the output significants structure.  By default,
	*  clusterSingles is false.
	*
	*  The significant tile structure is updated to contain the cluster 
	*  identification number and number of tiles in cluster
	*  \brief Density -based cluster finding.
	*  \param significants   Event list of significant tile properties
	*  \param distanceMetric Choice of metric to compute distance (see 
	*                        distance())
	*  \param durationInflation Multiplicative scale factor for duration 
	*                           (see distance())
	*  \param bandwidthInflation Multiplicative scale factor for bandwidth
	*                           (see distance())
	*  \param debugLevel Verbosity of debug output
	*/
      void wcluster(weventstack& significants, double clusterRadius=1.0, 
		    double clusterDensity=1.0, bool clusterSingles=false, 
		    const std::string& distanceMetric="",
		    double durationInflation=1.0, 
		    double bandwidthInflation=1.0, 
		    int debugLevel=1);


      /**  wcluster identifies clusters of related significant Q transform 
        *  tiles to improve the detection of gravitational-wave bursts that are
	*  extended in time and/or frequency.
	*
	*  <b> Hierarchical clustering <\b>
	*
	*  In hierarchical clustering, the distances returned by WDISTANCE are
	*  used by the linkage class to construct a hierarchical cluster tree 
	*  by successively linking together the next closest pair of tiles or 
	*  existing clusters.  Finally, clusters are identified by cutting the
	*  hieararchical cluster tree based on the specified criterion and
	*  threshold.
	*
	*  The cluster linkage method defines how intercluster distance is 
	*  computed, and may be any of the following methods.  By default,
	*  'single' linkage is assumed.
	*
	*  <table>
	*  <tr><td>'single'</td>
	*      <td> nearest distance between cluster members</td></tr>
	*  <tr><td>'complete'</td>
	*      <td>furthest distance between cluster members</td></tr>
	*  <tr><td>'average'</td>
	*      <td>average distance between cluster members</td></tr>
	*  <tr><td>'centroid'</td>
	*      <td>center of mass distance between clusters</td></tr>
	*  <tr><td>'ward'</td>
	*      <td>inner squared distance between clusters</td></tr>
	*  </table>
	*
	*  The cluster criterion specifies how to cut the hierarchical cluster 
	*  tree to form clusters, and may be either of the two methods 
	*  understood by the CLUSTER function.  By deafult, 'inconsistent' is 
	*  assumed.
	*  <table>
	*  <tr><td>'inconsistent'</td>
	*      <td>threshold on inconsistency between of cluster distance and
	*          average subcluster difference</td></tr>
	*  <tr><td>'distance'</td>
	*      <td>threshold on absolute distance between clusters</td></tr>
	*  </table>
	*
	*  The clusterThreshold corresponds to the specified clusterCriterion, 
	*  and has a default threshold of unity for both criteria.
	*/
      void wcluster(weventstack& significants, 
		    const std::string& clusterLinkage, 
		    const std::string& clusterCriterion, 
		    double clusterThreshold, 
		    const std::string& distanceMetric,
		    double durationInflation, double bandwidthInflation, 
		    int debugLevel);

      /** WSELECT selects statistically significant events from the set of 
       * statistically significant Q transform tiles. Events are defined by 
       * the properties of their most significant tile and are identified 
       * by exluding the less significant of any overlapping tiles.  The 
       * input significant tiles are first sorted by decreasing normalized 
       * energy.  Starting with the most significant tile, tiles are discarded 
       * if (they overlap with a more significant tile.  The remaining set of 
       * tiles comprises a minimal set of tiles that describes an event.
       * the results are left in the cuurent eventlist instance.
       *
       * The optional durationInflation and bandwidthInflation arguments are
       * multiplicative scale factors that are applied to the duration and 
       * bandwidth of significant tiles prior to testing for overlap.  If not 
       * specified, these parameters both default to unity such that the 
       * resulting tiles have unity time-frequency area.  The normalized 
       * energy of the resulting tiles are scaled by the product of the 
       * duration and bandwidth inflation factors to avoid over counting the 
       * total energy of clusters of tiles.  Likewise, the amplitude of the 
       * resulting tiles is scaled by the square root of the product of the
       * duration and bandwidth inflation factors.
       *
       * The optional maximumEvents argument provides a safety mechanism to 
       * limit the total number of events returned by WSELECT.  If this 
       * maximum number of events is exceeded, an overflow flag is set, only 
       * the maximumEvents most significant events are returned, and a warning
       * is issued if (debugLevel is set to unity or higher.  By default, 
       * maximumEvents is set to infinity and debugLevel is set to unity.
       *
       * WSELECT both expects and returns a cell array of Q transform event 
       * structures with one cell per channel.  The event structures contain 
       * the following required fields, which describe the properties of 
       * statistically significant tiles.  Additional fields such as amplitude, 
       * phase, or coherent transform properties are optional and are retained 
       * along with the required fields.
       *
       *   time                 center time of tile [gps seconds]
       *   frequency            center frequency of tile [Hz]
       *   duration             duration of tile [seconds]
       *   bandwidth            bandwidth of tile [Hz]
       *   normalizedEnergy     normalized energy of tile []
       *
       * The event list also contains the following flag, which indicates if 
       * the maximum number of significant tiles or significant events was 
       * exceeded.
       *
       * \param significants       event list tiles properties. This is assumed
       *                           to be ordered by decreasing normalizedEnergy.
       * \param durationInflation  multiplicative scale factor for duration
       * \param bandwidthInflation multiplicative scale factor for bandwidth
       * \param maximumEvents      maximum allowable number of events
       * \param debugLevel         verbosity of debug output
       */
      void wselect(const weventstack& significants, double durationInflation=1,
		   double bandwidthInflation=1.0, size_t maximumEvents=0, 
		   int debugLevel=1);

      /**  wthreshold identifies discrete Q transform coefficients whose 
       *  magnitudes exceed the threshold that approximately yields the 
       *  specified single channel false rate assuming ideal white noise.
       *  Each such tile is used to create an event which is added to the 
       *  list.
       *
       *  The user can focus on a subset of the times and frequencies available 
       *  in the transform data by specifying a desired range of central times,
       *  central frequencies, and Qs to threshold on.  Ranges should be 
       *  specified as a two component vector, consisting of a minimum and 
       *  maximum value. Alternatively, if (only a single Q is specified, 
       *  WTHRESHOLD is only applied to the time-frequency plane which has the 
       *  nearest value of Q in a logarithmic sense to the requested value.
       *
       *  To determine the range of central times to threshold on, WTHRESHOLD
       *  requires the start time of the transformed data in addition to a
       *  reference time and a relative time range.  Both the start time and
       *  reference time should be specified as absolute quantities, while the
       *  range of times to analyze should be specified relative to the 
       *  requested reference time.
       *
       *  By default, WTHRESHOLD is applied to all available frequencies and Qs,
       *  and the reference time and relative time range arguments are set to
       *  exclude data potentially corrupted by filter transients as identified 
       *  by the transient duration field of the tiling structure.  The default 
       *  value can be obtained for any argument by passing an empty vector.
       *
       *  The threshold is set to yield the specified false event rate when 
       *  applied to all available frequencies and Qs, and is not modified 
       *  to account for restricted ranges.  It is also only a rough estimate,
       *  and the result false event rate may vary significantly depending on 
       *  the quality of the data.
       *
       *  If provided, the optional analysisMode string is used by WTHRESHOLD 
       *  to determine which channels are signal channels, which channels are 
       *  null channels, and which channels to report results for.  For 
       *  coherent analysis modes, a desired white noise veto rate, squared 
       *  calibration uncertainy factor, and required signal correlation factor
       *  must also be specified.
       *
       *  The optional maximumSignificants argument provides a safety mechanism 
       *  to limit the total number of events returned by WTHRESHOLD.  If this 
       *  maximum number of significants is exceeded, the overflow flag is set, 
       *  only the maximumSignificants most significant tiles are returned, and 
       *  a warning is issued if (debugLevel is set to 1 or higher.  By default,
       *  maximumSignificants is set to 1e9-1 and debugLevel is set to unity.
       *
       *  \brief Generate list of significant tiles.
       *  \param signal         Input signal Q transform structure
       *  \param tiling         Discrete Q transform tiling structure from WTILE
       *  \param startTime      GPS start time of Q transformed data
       *  \param falseEventRate desired white noise false event rate [Hz]
       *  \param referenceTime  reference time for time range to threshold on
       *  \param timeRange      vector range of relative times to threshold on
       *  \param frequencyRange vector range of frequencies to threshold on
       *  \param qRange         scalar Q or vector range of Qs to threshold on
       *  \param maximumSignificants maximum allowable number of significant tiles
       *  \param analysisMode      string name of analysis mode to implement
       *  \param falseVetoRate     desired white noise veto rate [Hz]
       *  \param uncertaintyFactor squared calibration uncertainty factor
       *  \param correlationFactor fractional correlated energy threshold
       *  \param debugLevel        verbosity of debug output
       */
      void wthreshold(const wtransform& transforms, const wtile& tiling, 
		      const Time& startTime, double falseEventRate, 
		      const Time& referenceTime, const dble_vect& timeRange, 
		      const dble_vect& frequencyRange, const dble_vect& qRange, 
		      size_t maximumSignificants, 
		      const std::string& analysisMode="independent", 
		      double falseVetoRate= 0.0, double uncertaintyFactor=0.0, 
		      double correlationFactor=0.0, int debugLevel=1);

      /**  Write the specified fields of the event list to the output file.
       */
      void writeEvents(const wouttype& out, const str_vect& fields, 
		      const std::string& format) const;
 
   private:
      typedef  std::vector<weventlist> eventlist_vect;
      eventlist_vect _lists;
   };

   //====================================  Inline methods.
   inline const weventlist& 
   weventstack::operator[](int i) const {
      return _lists[i];
   }

   inline int 
   weventstack::numberOfChannels(void) const {
      return _lists.size();
   }

}

#endif // !defined(WPIPE_WEVENTLIST_HH)
